@@ -32,7 +32,7 @@
 #include "file.h"
 
 #ifndef	lint
-FILE_RCSID("@(#)$File: apprentice.c,v 1.173 2011/12/08 12:38:24 rrt Exp $")
+FILE_RCSID("@(#)$File: apprentice.c,v 1.181 2013/01/03 23:11:38 christos Exp $")
 #endif	/* lint */
 
 #include "magic.h"
@@ -71,6 +71,9 @@ FILE_RCSID("@(#)$File: apprentice.c,v 1.173 2011/12/08 12:38:24 rrt Exp $")
 #define MAP_FILE 0
 #endif
 
+#define ALLOC_CHUNK	(size_t)10
+#define ALLOC_INCR	(size_t)200
+
 struct magic_entry {
 	struct magic *mp;	
 	uint32_t cont_count;
@@ -86,15 +89,17 @@ private int getvalue(struct magic_set *ms, struct magic *, const char **, int);
 private int hextoint(int);
 private const char *getstr(struct magic_set *, struct magic *, const char *,
     int);
-private int parse(struct magic_set *, struct magic_entry **, uint32_t *,
-    const char *, size_t, int);
+private int parse(struct magic_set *, struct magic_entry *, const char *,
+    size_t, int);
 private void eatsize(const char **);
-private int apprentice_1(struct magic_set *, const char *, int, struct mlist *);
+private int apprentice_1(struct magic_set *, const char *, int);
 private size_t apprentice_magic_strength(const struct magic *);
 private int apprentice_sort(const void *, const void *);
 private void apprentice_list(struct mlist *, int );
 private int apprentice_load(struct magic_set *, struct magic **, uint32_t *,
     const char *, int);
+private struct mlist *mlist_alloc(void);
+private void mlist_free(struct mlist *);
 private void byteswap(struct magic *, uint32_t);
 private void bs1(struct magic *);
 private uint16_t swap2(uint16_t);
@@ -113,7 +118,7 @@ private int parse_strength(struct magic_set *, struct magic_entry *, const char
 private int parse_apple(struct magic_set *, struct magic_entry *, const char *);
 
 
-private size_t maxmagic = 0;
+private size_t maxmagic[MAGIC_SETS] = { 0 };
 private size_t magicsize = sizeof(struct magic);
 
 private const char usg_hdr[] = "cont\toffset\ttype\topcode\tmask\tvalue\tdesc";
@@ -212,7 +217,12 @@ static const struct type_tbl_s {
 	{ XX("ledouble"),	FILE_LEDOUBLE,		FILE_FMT_DOUBLE },
 	{ XX("leid3"),		FILE_LEID3,		FILE_FMT_NUM },
 	{ XX("beid3"),		FILE_BEID3,		FILE_FMT_NUM },
-	{ XX("indirect"),	FILE_INDIRECT,		FILE_FMT_NONE },
+	{ XX("indirect"),	FILE_INDIRECT,		FILE_FMT_NUM },
+	{ XX("qwdate"),		FILE_QWDATE,		FILE_FMT_STR },
+	{ XX("leqwdate"),	FILE_LEQWDATE,		FILE_FMT_STR },
+	{ XX("beqwdate"),	FILE_BEQWDATE,		FILE_FMT_STR },
+	{ XX("name"),		FILE_NAME,		FILE_FMT_STR },
+	{ XX("use"),		FILE_USE,		FILE_FMT_STR },
 	{ XX_NULL,		FILE_INVALID,		FILE_FMT_NONE },
 # undef XX
 # undef XX_NULL
@@ -250,77 +260,138 @@ init_file_tables(void)
 	}
 }
 
+private int
+add_mlist(struct mlist *mlp, struct magic *magic, uint32_t nmagic, int mapped)
+{
+	struct mlist *ml;
+
+	if ((ml = CAST(struct mlist *, malloc(sizeof(*ml)))) == NULL)
+		return -1;
+
+	ml->magic = magic;
+	ml->nmagic = nmagic;
+	ml->mapped = mapped;
+
+	mlp->prev->next = ml;
+	ml->prev = mlp->prev;
+	ml->next = mlp;
+	mlp->prev = ml;
+	return 0;
+}
+
 /*
  * Handle one file or directory.
  */
 private int
-apprentice_1(struct magic_set *ms, const char *fn, int action,
-    struct mlist *mlist)
+apprentice_1(struct magic_set *ms, const char *fn, int action)
 {
-	struct magic *magic = NULL;
-	uint32_t nmagic = 0;
+	struct magic *magic[MAGIC_SETS] = { NULL };
+	uint32_t nmagic[MAGIC_SETS] = { 0 };
 	struct mlist *ml;
 	int rv = -1;
 	int mapped;
+	size_t i;
 
 	if (magicsize != FILE_MAGICSIZE) {
 		file_error(ms, 0, "magic element size %lu != %lu",
-		    (unsigned long)sizeof(*magic),
+		    (unsigned long)sizeof(*magic[0]),
 		    (unsigned long)FILE_MAGICSIZE);
 		return -1;
 	}
 
 	if (action == FILE_COMPILE) {
-		rv = apprentice_load(ms, &magic, &nmagic, fn, action);
+		rv = apprentice_load(ms, magic, nmagic, fn, action);
 		if (rv != 0)
 			return -1;
-		rv = apprentice_compile(ms, &magic, &nmagic, fn);
-		free(magic);
+		rv = apprentice_compile(ms, magic, nmagic, fn);
 		return rv;
 	}
 
 #ifndef COMPILE_ONLY
-	if ((rv = apprentice_map(ms, &magic, &nmagic, fn)) == -1) {
+	if ((rv = apprentice_map(ms, magic, nmagic, fn)) < 0) {
+		if (rv == -2)
+			return -1;
 		if (ms->flags & MAGIC_CHECK)
 			file_magwarn(ms, "using regular magic file `%s'", fn);
-		rv = apprentice_load(ms, &magic, &nmagic, fn, action);
+		rv = apprentice_load(ms, magic, nmagic, fn, action);
 		if (rv != 0)
 			return -1;
 	}
 
 	mapped = rv;
 	     
-	if (magic == NULL) {
-		file_delmagic(magic, mapped, nmagic);
-		return -1;
+	for (i = 0; i < MAGIC_SETS; i++) {
+		if (magic[i] == NULL)
+			continue;
+		if (add_mlist(ms->mlist[i], magic[i], nmagic[i], mapped) == -1)
+		{
+			i = i == 1 ? 0 : 1;
+			file_delmagic(magic[i], mapped, nmagic[i]);
+			file_oomem(ms, sizeof(*ml));
+			return -1;
+		}
 	}
 
-	if ((ml = CAST(struct mlist *, malloc(sizeof(*ml)))) == NULL) {
-		file_delmagic(magic, mapped, nmagic);
-		file_oomem(ms, sizeof(*ml));
-		return -1;
-	}
-
-	ml->magic = magic;
-	ml->nmagic = nmagic;
-	ml->mapped = mapped;
-
-	mlist->prev->next = ml;
-	ml->prev = mlist->prev;
-	ml->next = mlist;
-	mlist->prev = ml;
-
 	if (action == FILE_LIST) {
-		printf("Binary patterns:\n");
-		apprentice_list(mlist, BINTEST);
-		printf("Text patterns:\n");
-		apprentice_list(mlist, TEXTTEST);
+		for (i = 0; i < MAGIC_SETS; i++) {
+			printf("Set %zu:\nBinary patterns:\n", i);
+			apprentice_list(ms->mlist[i], BINTEST);
+			printf("Text patterns:\n");
+			apprentice_list(ms->mlist[i], TEXTTEST);
+		}
 	}
 	
 	return 0;
 #endif /* COMPILE_ONLY */
 }
 
+protected void
+file_ms_free(struct magic_set *ms)
+{
+	size_t i;
+	if (ms == NULL)
+		return;
+	for (i = 0; i < MAGIC_SETS; i++)
+		mlist_free(ms->mlist[i]);
+	free(ms->o.pbuf);
+	free(ms->o.buf);
+	free(ms->c.li);
+	free(ms);
+}
+
+protected struct magic_set *
+file_ms_alloc(int flags)
+{
+	struct magic_set *ms;
+	size_t i, len;
+
+	if ((ms = CAST(struct magic_set *, calloc((size_t)1,
+	    sizeof(struct magic_set)))) == NULL)
+		return NULL;
+
+	if (magic_setflags(ms, flags) == -1) {
+		errno = EINVAL;
+		goto free;
+	}
+
+	ms->o.buf = ms->o.pbuf = NULL;
+	len = (ms->c.len = 10) * sizeof(*ms->c.li);
+
+	if ((ms->c.li = CAST(struct level_info *, malloc(len))) == NULL)
+		goto free;
+
+	ms->event_flags = 0;
+	ms->error = -1;
+	for (i = 0; i < MAGIC_SETS; i++)
+		ms->mlist[i] = NULL;
+	ms->file = "unknown";
+	ms->line = 0;
+	return ms;
+free:
+	free(ms);
+	return NULL;
+}
+
 protected void
 file_delmagic(struct magic *p, int type, size_t entries)
 {
@@ -348,51 +419,108 @@ file_delmagic(struct magic *p, int type, size_t entries)
 	}
 }
 
+private struct mlist *
+mlist_alloc(void)
+{
+	struct mlist *mlist;
+	if ((mlist = CAST(struct mlist *, malloc(sizeof(*mlist)))) == NULL) {
+		return NULL;
+	}
+	mlist->next = mlist->prev = mlist;
+	return mlist;
+}
+
+private void
+mlist_free(struct mlist *mlist)
+{
+	struct mlist *ml;
+
+	if (mlist == NULL)
+		return;
+
+	for (ml = mlist->next; ml != mlist;) {
+		struct mlist *next = ml->next;
+		struct magic *mg = ml->magic;
+		file_delmagic(mg, ml->mapped, ml->nmagic);
+		free(ml);
+		ml = next;
+	}
+	free(ml);
+}
+
 /* const char *fn: list of magic files and directories */
-protected struct mlist *
+protected int
 file_apprentice(struct magic_set *ms, const char *fn, int action)
 {
 	char *p, *mfn;
 	int file_err, errs = -1;
-	struct mlist *mlist;
+	size_t i;
 
 	if ((fn = magic_getpath(fn, action)) == NULL)
-		return NULL;
+		return -1;
 
 	init_file_tables();
 
 	if ((mfn = strdup(fn)) == NULL) {
 		file_oomem(ms, strlen(fn));
-		return NULL;
+		return -1;
+	}
+
+	for (i = 0; i < MAGIC_SETS; i++) {
+		mlist_free(ms->mlist[i]);
+		if ((ms->mlist[i] = mlist_alloc()) == NULL) {
+			file_oomem(ms, sizeof(*ms->mlist[i]));
+			if (i != 0) {
+				--i;
+				do
+					mlist_free(ms->mlist[i]);
+				while (i != 0);
+			}
+			free(mfn);
+			return -1;
+		}
 	}
 	fn = mfn;
 
-	if ((mlist = CAST(struct mlist *, malloc(sizeof(*mlist)))) == NULL) {
-		free(mfn);
-		file_oomem(ms, sizeof(*mlist));
-		return NULL;
-	}
-	mlist->next = mlist->prev = mlist;
-
 	while (fn) {
 		p = strchr(fn, PATHSEP);
 		if (p)
 			*p++ = '\0';
 		if (*fn == '\0')
 			break;
-		file_err = apprentice_1(ms, fn, action, mlist);
+		file_err = apprentice_1(ms, fn, action);
 		errs = MAX(errs, file_err);
 		fn = p;
 	}
-	if (errs == -1) {
-		free(mfn);
-		free(mlist);
-		mlist = NULL;
-		file_error(ms, 0, "could not find any magic files!");
-		return NULL;
-	}
+
 	free(mfn);
-	return mlist;
+
+	if (errs == -1) {
+		for (i = 0; i < MAGIC_SETS; i++) {
+			mlist_free(ms->mlist[i]);
+			ms->mlist[i] = NULL;
+		}
+		file_error(ms, 0, "could not find any magic files!");
+		return -1;
+	}
+
+	if (action == FILE_LOAD)
+		return 0;
+
+	for (i = 0; i < MAGIC_SETS; i++) {
+		mlist_free(ms->mlist[i]);
+		ms->mlist[i] = NULL;
+	}
+
+	switch (action) {
+	case FILE_COMPILE:
+	case FILE_CHECK:
+	case FILE_LIST:
+		return 0;
+	default:
+		file_error(ms, 0, "Invalid action %d", action);
+		return -1;
+	}
 }
 
 /*
@@ -465,12 +593,20 @@ apprentice_magic_strength(const struct magic *m)
 	case FILE_QLDATE:
 	case FILE_LEQLDATE:
 	case FILE_BEQLDATE:
+	case FILE_QWDATE:
+	case FILE_LEQWDATE:
+	case FILE_BEQWDATE:
 	case FILE_DOUBLE:
 	case FILE_BEDOUBLE:
 	case FILE_LEDOUBLE:
 		val += 8 * MULT;
 		break;
 
+	case FILE_INDIRECT:
+	case FILE_NAME:
+	case FILE_USE:
+		break;
+
 	default:
 		val = 0;
 		(void)fprintf(stderr, "Bad type %d\n", m->type);
@@ -617,6 +753,9 @@ set_test_type(struct magic *mstart, struct magic *m)
 	case FILE_QLDATE:
 	case FILE_LEQLDATE:
 	case FILE_BEQLDATE:
+	case FILE_QWDATE:
+	case FILE_LEQWDATE:
+	case FILE_BEQWDATE:
 	case FILE_FLOAT:
 	case FILE_BEFLOAT:
 	case FILE_LEFLOAT:
@@ -664,16 +803,41 @@ set_test_type(struct magic *mstart, struct magic *m)
 	}
 }
 
+private int
+addentry(struct magic_set *ms, struct magic_entry *me,
+   struct magic_entry **mentry, uint32_t *mentrycount)
+{
+	size_t i = me->mp->type == FILE_NAME ? 1 : 0;
+	if (mentrycount[i] == maxmagic[i]) {
+		struct magic_entry *mp;
+
+		maxmagic[i] += ALLOC_INCR;
+		if ((mp = CAST(struct magic_entry *,
+		    realloc(mentry[i], sizeof(*mp) * maxmagic[i]))) ==
+		    NULL) {
+			file_oomem(ms, sizeof(*mp) * maxmagic[i]);
+			return -1;
+		}
+		(void)memset(&mp[mentrycount[i]], 0, sizeof(*mp) *
+		    ALLOC_INCR);
+		mentry[i] = mp;
+	}
+	mentry[i][mentrycount[i]++] = *me;
+	memset(me, 0, sizeof(*me));
+	return 0;
+}
+
 /*
  * Load and parse one file.
  */
 private void
 load_1(struct magic_set *ms, int action, const char *fn, int *errs,
-   struct magic_entry **marray, uint32_t *marraycount)
+   struct magic_entry **mentry, uint32_t *mentrycount)
 {
 	size_t lineno = 0, llen = 0;
 	char *line = NULL;
 	ssize_t len;
+	struct magic_entry me;
 
 	FILE *f = fopen(ms->file = fn, "r");
 	if (f == NULL) {
@@ -684,6 +848,7 @@ load_1(struct magic_set *ms, int action, const char *fn, int *errs,
 		return;
 	}
 
+	memset(&me, 0, sizeof(me));
 	/* read and parse this file */
 	for (ms->line = 1; (len = getline(&line, &llen, f)) != -1;
 	    ms->line++) {
@@ -713,15 +878,14 @@ load_1(struct magic_set *ms, int action, const char *fn, int *errs,
 					(*errs)++;
 					continue;
 				}
-				if (*marraycount == 0) {
+				if (me.mp == NULL) {
 					file_error(ms, 0,
 					    "No current entry for :!%s type",
 						bang[i].name);
 					(*errs)++;
 					continue;
 				}
-				if ((*bang[i].fun)(ms, 
-				    &(*marray)[*marraycount - 1],
+				if ((*bang[i].fun)(ms, &me,
 				    line + bang[i].len + 2) != 0) {
 					(*errs)++;
 					continue;
@@ -730,12 +894,21 @@ load_1(struct magic_set *ms, int action, const char *fn, int *errs,
 			}
 			/*FALLTHROUGH*/
 		default:
-			if (parse(ms, marray, marraycount, line, lineno,
-			    action) != 0)
+		again:
+			switch (parse(ms, &me, line, lineno, action)) {
+			case 0:
+				continue;
+			case 1:
+				(void)addentry(ms, &me, mentry, mentrycount);
+				goto again;
+			default:
 				(*errs)++;
-			break;
+				break;
+			}
 		}
 	}
+	if (me.mp)
+		(void)addentry(ms, &me, mentry, mentrycount);
 	free(line);
 	(void)fclose(f);
 }
@@ -750,14 +923,107 @@ cmpstrp(const void *p1, const void *p2)
         return strcmp(*(char *const *)p1, *(char *const *)p2);
 }
 
+
+private uint32_t
+set_text_binary(struct magic_set *ms, struct magic_entry *me, uint32_t nme,
+    uint32_t starttest)
+{
+	static const char text[] = "text";
+	static const char binary[] = "binary";
+	static const size_t len = sizeof(text);
+
+	uint32_t i = starttest;
+
+	do {
+		set_test_type(me[starttest].mp, me[i].mp);
+		if ((ms->flags & MAGIC_DEBUG) == 0)
+			continue;
+		(void)fprintf(stderr, "%s%s%s: %s\n",
+		    me[i].mp->mimetype,
+		    me[i].mp->mimetype[0] == '\0' ? "" : "; ",
+		    me[i].mp->desc[0] ? me[i].mp->desc : "(no description)",
+		    me[i].mp->flag & BINTEST ? binary : text);
+		if (me[i].mp->flag & BINTEST) {
+			char *p = strstr(me[i].mp->desc, text);
+			if (p && (p == me[i].mp->desc ||
+			    isspace((unsigned char)p[-1])) &&
+			    (p + len - me[i].mp->desc == MAXstring
+			    || (p[len] == '\0' ||
+			    isspace((unsigned char)p[len]))))
+				(void)fprintf(stderr, "*** Possible "
+				    "binary test for text type\n");
+		}
+	} while (++i < nme && me[i].mp->cont_level != 0);
+	return i;
+}
+
+private void
+set_last_default(struct magic_set *ms, struct magic_entry *me, uint32_t nme)
+{
+	uint32_t i;
+	for (i = 0; i < nme; i++) {
+		if (me[i].mp->cont_level == 0 &&
+		    me[i].mp->type == FILE_DEFAULT) {
+			while (++i < nme)
+				if (me[i].mp->cont_level == 0)
+					break;
+			if (i != nme) {
+				/* XXX - Ugh! */
+				ms->line = me[i].mp->lineno;
+				file_magwarn(ms,
+				    "level 0 \"default\" did not sort last");
+			}
+			return;					    
+		}
+	}
+}
+
+private int
+coalesce_entries(struct magic_set *ms, struct magic_entry *me, uint32_t nme,
+    struct magic **ma, uint32_t *nma)
+{
+	uint32_t i, mentrycount = 0;
+	size_t slen;
+
+	for (i = 0; i < nme; i++)
+		mentrycount += me[i].cont_count;
+
+	slen = sizeof(**ma) * mentrycount;
+	if ((*ma = CAST(struct magic *, malloc(slen))) == NULL) {
+		file_oomem(ms, slen);
+		return -1;
+	}
+
+	mentrycount = 0;
+	for (i = 0; i < nme; i++) {
+		(void)memcpy(*ma + mentrycount, me[i].mp,
+		    me[i].cont_count * sizeof(**ma));
+		mentrycount += me[i].cont_count;
+	}
+	*nma = mentrycount;
+	return 0;
+}
+
+private void
+magic_entry_free(struct magic_entry *me, uint32_t nme)
+{
+	uint32_t i;
+	if (me == NULL)
+		return;
+	for (i = 0; i < nme; i++)
+		free(me[i].mp);
+	free(me);
+}
+
 private int
 apprentice_load(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
     const char *fn, int action)
 {
 	int errs = 0;
-	struct magic_entry *marray;
-	uint32_t marraycount, i, mentrycount = 0, starttest;
-	size_t slen, files = 0, maxfiles = 0;
+	struct magic_entry *mentry[MAGIC_SETS] = { NULL };
+	uint32_t mentrycount[MAGIC_SETS] = { 0 };
+	uint32_t i, j;
+	size_t files = 0, maxfiles = 0;
 	char **filearr = NULL, *mfn;
 	struct stat st;
 	DIR *dir;
@@ -765,14 +1031,6 @@ apprentice_load(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
 
 	ms->flags |= MAGIC_CHECK;	/* Enable checks for parsed files */
 
-        maxmagic = MAXMAGIS;
-	if ((marray = CAST(struct magic_entry *, calloc(maxmagic,
-	    sizeof(*marray)))) == NULL) {
-		file_oomem(ms, maxmagic * sizeof(*marray));
-		return -1;
-	}
-	marraycount = 0;
-
 	/* print silly verbose header for USG compat. */
 	if (action == FILE_CHECK)
 		(void)fprintf(stderr, "%s\n", usg_hdr);
@@ -814,100 +1072,57 @@ apprentice_load(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
 		closedir(dir);
 		qsort(filearr, files, sizeof(*filearr), cmpstrp);
 		for (i = 0; i < files; i++) {
-			load_1(ms, action, filearr[i], &errs, &marray,
-			    &marraycount);
+			load_1(ms, action, filearr[i], &errs, mentry,
+			    mentrycount);
 			free(filearr[i]);
 		}
 		free(filearr);
 	} else
-		load_1(ms, action, fn, &errs, &marray, &marraycount);
+		load_1(ms, action, fn, &errs, mentry, mentrycount);
 	if (errs)
 		goto out;
 
-	/* Set types of tests */
-	for (i = 0; i < marraycount; ) {
-		if (marray[i].mp->cont_level != 0) {
-			i++;
-			continue;
-		}
-
-		starttest = i;
-		do {
-			static const char text[] = "text";
-			static const char binary[] = "binary";
-			static const size_t len = sizeof(text);
-			set_test_type(marray[starttest].mp, marray[i].mp);
-			if ((ms->flags & MAGIC_DEBUG) == 0)
+	for (j = 0; j < MAGIC_SETS; j++) {
+		/* Set types of tests */
+		for (i = 0; i < mentrycount[j]; ) {
+			if (mentry[j][i].mp->cont_level != 0) {
+				i++;
 				continue;
-			(void)fprintf(stderr, "%s%s%s: %s\n",
-			    marray[i].mp->mimetype,
-			    marray[i].mp->mimetype[0] == '\0' ? "" : "; ",
-			    marray[i].mp->desc[0] ? marray[i].mp->desc :
-			    "(no description)",
-			    marray[i].mp->flag & BINTEST ? binary : text);
-			if (marray[i].mp->flag & BINTEST) {
-				char *p = strstr(marray[i].mp->desc, text);
-				if (p && (p == marray[i].mp->desc ||
-				    isspace((unsigned char)p[-1])) &&
-				    (p + len - marray[i].mp->desc == 
-				    MAXstring || (p[len] == '\0' ||
-				    isspace((unsigned char)p[len]))))
-					(void)fprintf(stderr, "*** Possible "
-					    "binary test for text type\n");
 			}
-		} while (++i < marraycount && marray[i].mp->cont_level != 0);
-	}
+			i = set_text_binary(ms, mentry[j], mentrycount[j], i);
+		}
+		qsort(mentry[j], mentrycount[j], sizeof(*mentry[j]),
+		    apprentice_sort);
 
-	qsort(marray, marraycount, sizeof(*marray), apprentice_sort);
+		/*
+		 * Make sure that any level 0 "default" line is last
+		 * (if one exists).
+		 */
+		set_last_default(ms, mentry[j], mentrycount[j]);
 
-	/*
-	 * Make sure that any level 0 "default" line is last (if one exists).
-	 */
-	for (i = 0; i < marraycount; i++) {
-		if (marray[i].mp->cont_level == 0 &&
-		    marray[i].mp->type == FILE_DEFAULT) {
-			while (++i < marraycount)
-				if (marray[i].mp->cont_level == 0)
-					break;
-			if (i != marraycount) {
-				/* XXX - Ugh! */
-				ms->line = marray[i].mp->lineno;
-				file_magwarn(ms,
-				    "level 0 \"default\" did not sort last");
-			}
-			break;					    
+		/* coalesce per file arrays into a single one */
+		if (coalesce_entries(ms, mentry[j], mentrycount[j],
+		    &magicp[j], &nmagicp[j]) == -1) {
+			errs++;
+			goto out;
 		}
 	}
 
-	for (i = 0; i < marraycount; i++)
-		mentrycount += marray[i].cont_count;
-
-	slen = sizeof(**magicp) * mentrycount;
-	if ((*magicp = CAST(struct magic *, malloc(slen))) == NULL) {
-		file_oomem(ms, slen);
-		errs++;
-		goto out;
-	}
-
-	mentrycount = 0;
-	for (i = 0; i < marraycount; i++) {
-		(void)memcpy(*magicp + mentrycount, marray[i].mp,
-		    marray[i].cont_count * sizeof(**magicp));
-		mentrycount += marray[i].cont_count;
-	}
 out:
-	for (i = 0; i < marraycount; i++)
-		free(marray[i].mp);
-	free(marray);
-	if (errs) {
-		*magicp = NULL;
-		*nmagicp = 0;
-		return errs;
-	} else {
-		*nmagicp = mentrycount;
-		return 0;
-	}
+	for (j = 0; j < MAGIC_SETS; j++)
+		magic_entry_free(mentry[j], mentrycount[j]);
 
+	if (errs) {
+		for (j = 0; j < MAGIC_SETS; j++) {
+			if (magicp[j]) {
+				free(magicp[j]);
+				magicp[j] = NULL;
+			}
+			nmagicp[j] = 0;
+		}
+		return errs;
+	}
+	return 0;
 }
 
 /*
@@ -953,10 +1168,13 @@ file_signextend(struct magic_set *ms, struct magic *m, uint64_t v)
 		case FILE_LEQUAD:
 		case FILE_QDATE:
 		case FILE_QLDATE:
+		case FILE_QWDATE:
 		case FILE_BEQDATE:
 		case FILE_BEQLDATE:
+		case FILE_BEQWDATE:
 		case FILE_LEQDATE:
 		case FILE_LEQLDATE:
+		case FILE_LEQWDATE:
 		case FILE_DOUBLE:
 		case FILE_BEDOUBLE:
 		case FILE_LEDOUBLE:
@@ -970,6 +1188,8 @@ file_signextend(struct magic_set *ms, struct magic *m, uint64_t v)
 		case FILE_SEARCH:
 		case FILE_DEFAULT:
 		case FILE_INDIRECT:
+		case FILE_NAME:
+		case FILE_USE:
 			break;
 		default:
 			if (ms->flags & MAGIC_CHECK)
@@ -1139,14 +1359,13 @@ check_cond(struct magic_set *ms, int cond, uint32_t cont_level)
  * parse one line from magic file, put into magic[index++] if valid
  */
 private int
-parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp, 
-    const char *line, size_t lineno, int action)
+parse(struct magic_set *ms, struct magic_entry *me, const char *line,
+    size_t lineno, int action)
 {
 #ifdef ENABLE_CONDITIONALS
 	static uint32_t last_cont_level = 0;
 #endif
 	size_t i;
-	struct magic_entry *me;
 	struct magic *m;
 	const char *l = line;
 	char *t;
@@ -1165,16 +1384,11 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 			return -1;
 	last_cont_level = cont_level;
 #endif
-
-#define ALLOC_CHUNK	(size_t)10
-#define ALLOC_INCR	(size_t)200
-
 	if (cont_level != 0) {
-		if (*nmentryp == 0) {
+		if (me->mp == NULL) {
 			file_error(ms, 0, "No current entry for continuation");
 			return -1;
 		}
-		me = &(*mentryp)[*nmentryp - 1];
 		if (me->cont_count == me->max_count) {
 			struct magic *nm;
 			size_t cnt = me->max_count + ALLOC_CHUNK;
@@ -1190,31 +1404,15 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 		(void)memset(m, 0, sizeof(*m));
 		m->cont_level = cont_level;
 	} else {
-		if (*nmentryp == maxmagic) {
-			struct magic_entry *mp;
-
-			maxmagic += ALLOC_INCR;
-			if ((mp = CAST(struct magic_entry *,
-			    realloc(*mentryp, sizeof(*mp) * maxmagic))) ==
-			    NULL) {
-				file_oomem(ms, sizeof(*mp) * maxmagic);
-				return -1;
-			}
-			(void)memset(&mp[*nmentryp], 0, sizeof(*mp) *
-			    ALLOC_INCR);
-			*mentryp = mp;
+		static const size_t len = sizeof(*m) * ALLOC_CHUNK;
+		if (me->mp != NULL)
+			return 1;
+		if ((m = CAST(struct magic *, malloc(len))) == NULL) {
+			file_oomem(ms, len);
+			return -1;
 		}
-		me = &(*mentryp)[*nmentryp];
-		if (me->mp == NULL) {
-			size_t len = sizeof(*m) * ALLOC_CHUNK;
-			if ((m = CAST(struct magic *, malloc(len))) == NULL) {
-				file_oomem(ms, len);
-				return -1;
-			}
-			me->mp = m;
-			me->max_count = ALLOC_CHUNK;
-		} else
-			m = me->mp;
+		me->mp = m;
+		me->max_count = ALLOC_CHUNK;
 		(void)memset(m, 0, sizeof(*m));
 		m->factor_op = FILE_FACTOR_OP_NONE;
 		m->cont_level = 0;
@@ -1344,7 +1542,7 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 	EATAB;
 #endif
 
-	if (*l == 'u') {
+	if (*l == 'u' && (l[1] != 's' || l[2] != 'e')) {
 		++l;
 		m->flag |= UNSIGNED;
 	}
@@ -1422,6 +1620,9 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 				case CHAR_TEXTTEST:
 					m->str_flags |= STRING_TEXTTEST;
 					break;
+				case CHAR_TRIM:
+					m->str_flags |= STRING_TRIM;
+					break;
 				case CHAR_PSTRING_1_LE:
 					if (m->type != FILE_PSTRING)
 						goto bad;
@@ -1564,8 +1765,6 @@ parse(struct magic_set *ms, struct magic_entry **mentryp, uint32_t *nmentryp,
 	}
 #endif
 	m->mimetype[0] = '\0';		/* initialise MIME type to none */
-	if (m->cont_level == 0)
-		++(*nmentryp);		/* make room for next */
 	return 0;
 }
 
@@ -1887,6 +2086,8 @@ getvalue(struct magic_set *ms, struct magic *m, const char **p, int action)
 	case FILE_PSTRING:
 	case FILE_REGEX:
 	case FILE_SEARCH:
+	case FILE_NAME:
+	case FILE_USE:
 		*p = getstr(ms, m, *p, action == FILE_COMPILE);
 		if (*p == NULL) {
 			if (ms->flags & MAGIC_CHECK)
@@ -2200,14 +2401,16 @@ private int
 apprentice_map(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
     const char *fn)
 {
-	int fd;
+	int fd, ret;
 	struct stat st;
 	uint32_t *ptr;
-	uint32_t version;
+	uint32_t version, entries, nentries;
 	int needsbyteswap;
 	char *dbname = NULL;
 	void *mm = NULL;
+	size_t i;
 
+	ret = -1;
 	dbname = mkdbname(ms, fn, 0);
 	if (dbname == NULL)
 		goto error2;
@@ -2242,10 +2445,9 @@ apprentice_map(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
 	}
 #define RET	1
 #endif
-	*magicp = CAST(struct magic *, mm);
 	(void)close(fd);
 	fd = -1;
-	ptr = (uint32_t *)(void *)*magicp;
+	ptr = CAST(uint32_t *, mm);
 	if (*ptr != MAGICNO) {
 		if (swap4(*ptr) != MAGICNO) {
 			file_error(ms, 0, "bad magic in `%s'", dbname);
@@ -2258,18 +2460,39 @@ apprentice_map(struct magic_set *ms, struct magic **magicp, uint32_t *nmagicp,
 		version = swap4(ptr[1]);
 	else
 		version = ptr[1];
+	ret = -2;
 	if (version != VERSIONNO) {
 		file_error(ms, 0, "File %s supports only version %d magic "
 		    "files. `%s' is version %d", VERSION,
 		    VERSIONNO, dbname, version);
 		goto error1;
 	}
-	*nmagicp = (uint32_t)(st.st_size / sizeof(struct magic));
-	if (*nmagicp > 0)
-		(*nmagicp)--;
-	(*magicp)++;
+	entries = (uint32_t)(st.st_size / sizeof(struct magic));
+	if ((off_t)(entries * sizeof(struct magic)) != st.st_size) {
+		file_error(ms, 0, "Size of `%s' %llu is not a multiple of %zu",
+		    dbname, (unsigned long long)st.st_size,
+		    sizeof(struct magic));
+		goto error1;
+	}
+	magicp[0] = CAST(struct magic *, mm) + 1;
+	nentries = 0;
+	for (i = 0; i < MAGIC_SETS; i++) {
+		if (needsbyteswap)
+			nmagicp[i] = swap4(ptr[i + 2]);
+		else
+			nmagicp[i] = ptr[i + 2];
+		if (i != MAGIC_SETS - 1)
+			magicp[i + 1] = magicp[i] + nmagicp[i];
+		nentries += nmagicp[i];
+	}
+	if (entries != nentries + 1) {
+		file_error(ms, 0, "Inconsistent entries in `%s' %u != %u",
+		    dbname, entries, nentries + 1);
+		goto error1;
+	}
 	if (needsbyteswap)
-		byteswap(*magicp, *nmagicp);
+		for (i = 0; i < MAGIC_SETS; i++)
+			byteswap(magicp[i], nmagicp[i]);
 	free(dbname);
 	return RET;
 
@@ -2283,12 +2506,14 @@ error1:
 		free(mm);
 #endif
 	} else {
-		*magicp = NULL;
-		*nmagicp = 0;
+		for (i = 0; i < MAGIC_SETS; i++) {
+			magicp[i] = NULL;
+			nmagicp[i] = 0;
+		}
 	}
 error2:
 	free(dbname);
-	return -1;
+	return ret;
 }
 
 private const uint32_t ar[] = {
@@ -2301,16 +2526,19 @@ private int
 apprentice_compile(struct magic_set *ms, struct magic **magicp,
     uint32_t *nmagicp, const char *fn)
 {
+	static size_t nm = sizeof(*nmagicp) * MAGIC_SETS;
 	int fd = -1;
 	char *dbname;
 	int rv = -1;
+	uint32_t i;
 
 	dbname = mkdbname(ms, fn, 1);
 
 	if (dbname == NULL) 
 		goto out;
 
-	if ((fd = open(dbname, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0644)) == -1) {
+	if ((fd = open(dbname, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, 0644)) == -1) 
+	{
 		file_error(ms, errno, "cannot open `%s'", dbname);
 		goto out;
 	}
@@ -2320,16 +2548,25 @@ apprentice_compile(struct magic_set *ms, struct magic **magicp,
 		goto out;
 	}
 
+	if (write(fd, nmagicp, nm) != (ssize_t)nm) {
+		file_error(ms, errno, "error writing `%s'", dbname);
+		goto out;
+	}
+
+	assert(nm + sizeof(ar) < sizeof(struct magic));
+
 	if (lseek(fd, (off_t)sizeof(struct magic), SEEK_SET)
 	    != sizeof(struct magic)) {
 		file_error(ms, errno, "error seeking `%s'", dbname);
 		goto out;
 	}
 
-	if (write(fd, *magicp, (sizeof(struct magic) * *nmagicp)) 
-	    != (ssize_t)(sizeof(struct magic) * *nmagicp)) {
-		file_error(ms, errno, "error writing `%s'", dbname);
-		goto out;
+	for (i = 0; i < MAGIC_SETS; i++) {
+		if (write(fd, magicp[i], (sizeof(struct magic) * nmagicp[i])) 
+		    != (ssize_t)(sizeof(struct magic) * nmagicp[i])) {
+			    file_error(ms, errno, "error writing `%s'", dbname);
+			    goto out;
+		}
 	}
 
 	if (fd != -1)
@@ -2524,3 +2761,30 @@ file_pstring_get_length(const struct magic *m, const char *s)
 
 	return len;
 }
+
+protected int
+file_magicfind(struct magic_set *ms, const char *name, struct mlist *v)
+{
+	uint32_t i, j;
+	struct mlist *mlist, *ml;
+
+	mlist = ms->mlist[1];
+
+	for (ml = mlist->next; ml != mlist; ml = ml->next) {
+		struct magic *ma = ml->magic;
+		uint32_t nma = ml->nmagic;
+		for (i = 0; i < nma; i++) {
+			if (ma[i].type != FILE_NAME)
+				continue;
+			if (strcmp(ma[i].value.s, name) == 0) {
+				v->magic = &ma[i];
+				for (j = i + 1; j < nma; j++)
+				    if (ma[j].cont_level == 0)
+					    break;
+				v->nmagic = j - i;
+				return 0;
+			}
+		}
+	}
+	return -1;
+}
